#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# (C) British Crown copyright. The Met Office.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""Script to clip the input cube's data to be between the specified values"""

from improver import cli


@cli.clizefy
@cli.with_output
def process(
    cube: cli.inputcube, *, min_value: float = None, max_value: float = None,
):
    """Clip the data in the input cube such that any data above max_value is set equal to
     max_value and any data below min_value is set equal to min_value.

    Args:
        cube (iris.cube.Cube):
            A Cube whose data will be clipped. This can be a cube of spot or gridded data.
        max_value (float):
            If specified any data in cube that is above max_value will be set equal to
            max_value.
        min_value (float):
            If specified any data in cube that is below min_value will be set equal to
            min_value.
    Returns:
        iris.cube.Cube:
            A cube with the same metadata as the input cube but with the data clipped such
            that any data above max_value is set equal to max_value and any data below
            min_value is set equal to min_value.
    """
    from numpy import clip

    cube.data = clip(cube.data, min_value, max_value)

    return cube
