# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# (C) British Crown copyright. The Met Office.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""Script to generate freezing rain probabilities."""

from improver import cli


@cli.clizefy
@cli.with_output
def process(*cubes: cli.inputcube, model_id_attr: str = None):
    """
    Calculates a probability of freezing-rain near the ground using rain, sleet,
    and temperature probabilities.

    P(freezing_rain rate or accumulation > threshold) = (
        (P(rain rate or accumulation >  threshold) +
         P(sleet rate or accumulation >  threshold)) * P(temperature < 0C)

    If the input data is multi-realization, the realization coordinate is
    collapsed as part of this calculation to yield an ensemble average
    probability of freezing rain.

    Args:
        cubes (iris.cube.CubeList or list):
            Contains cubes of rain, sleet, and temperature probabilities. The
            precipitation cubes may be either rates or accumulations. The
            temperature should be a surface or screen temperature. It may be a
            period minimum, with a period that matches the precipitation
            accumulation period, or an instantaneous temperature if using
            precipitation rates.
        model_id_attr (str):
            Name of the attribute used to identify the source model for
            blending.

    Returns:
        iris.cube.Cube:
            A cube of freezing rain rate or accumulation probabilities.

    """
    from iris.cube import CubeList

    from improver.precipitation_type.freezing_rain import FreezingRain

    return FreezingRain(model_id_attr=model_id_attr)(CubeList(cubes))
